#!/usr/bin/env python3
# coding=utf-8

"""Reduction from DQBF to effective propositional fragment of FOL."""

__author__ = "Leander Tentrup <tentrup@cs.uni-saarland.de>"
__copyright__ = "Copyright 2014, Saarland University"

import sys, os, re, copy

from lex import *
from parse import *
from pyeda.inter import *
from dqbf_solver import PyEDAVisitor

CNF_OUT = True

def printUsage():
    print('Usage: %s ["DQBF string" | -F file_name | stdin]' % sys.argv[0])
    exit(1)

def convert_var(var):
    reg_expr = re.compile(r"([a-zA-Z]+)(\d+)")
    result = reg_expr.match(var)
    if result:
        return exprvar(result.groups()[0], int(result.groups()[1]))
    else:
        return exprvar(var)

def var2str(var, universal=False):
    string_representation = var.name
    if universal:
        string_representation = string_representation.upper()
    if var.indices:
        string_representation += str(var.indices[0])
    return string_representation

def transform_formula_to_fof(formula):
    if isinstance(formula, ExprVariable):
        if formula in univ:
            return var2str(formula, universal=True)
        else:
            depends = ['{}'.format(var2str(v, universal=True)) for v in exvar_scope[formula]]
            return '{}({})'.format(var2str(formula), ','.join(depends))
    elif isinstance(formula, ExprComplement):
        return var2str(~formula, formula in univ)
    elif isinstance(formula, ExprNot):
        return "~" + transform_formula_to_fof(formula.arg)
    elif isinstance(formula, ExprOrAnd):
        operands = [transform_formula_to_fof(f) for f in formula.args]
        if isinstance(formula, ExprOr):
            return '({})'.format(' | '.join(operands))
        else:
            return '({})'.format(' & '.join(operands))
    
    elif isinstance(formula, ExprNor) or isinstance(formula, ExprNand):
        assert len(formula.args) == 2
        operands = [transform_formula_to_fof(f) for f in formula.args]
        if isinstance(formula, ExprNor):
            return '({})'.format(' ~| '.join(operands))
        else:
            return '({})'.format(' ~& '.join(operands))
            
    elif isinstance(formula, ExprExclusive):
        operands = [transform_formula_to_fof(f) for f in formula.args]
        if isinstance(formula, ExprXor):
            return '({})'.format(' <~> '.join(operands))
        else:
            return '({})'.format(' <=> '.join(operands))
            
    elif isinstance(formula, ExprEqual) or isinstance(formula, ExprUnequal):
        assert len(formula.args) == 2
        operands = [transform_formula_to_fof(f) for f in formula.args]
        if isinstance(formula, ExprUnequal):
            return '({})'.format(' <~> '.join(operands))
        else:
            return '({})'.format(' <=> '.join(operands))

    elif isinstance(formula, ExprImplies):
        assert len(formula.args) == 2
        operands = [transform_formula_to_fof(f) for f in formula.args]
        return '({})'.format(' => '.join(operands))

if __name__ == "__main__":
    
    #sys.setrecursionlimit(100000)
    
    process_stdin = False
    if not sys.stdin.isatty():
        process_stdin = True
        dqbf_string = sys.stdin.read()
        
        if len(sys.argv) != 1:
            printUsage()
    
    elif len(sys.argv) == 2:
        dqbf_string = sys.argv[1]
    elif len(sys.argv) == 3 and sys.argv[1] == '-F':
        file_handle = open(sys.argv[2], 'r')
        dqbf_string = file_handle.read()
    else:
        printUsage()
    
    parse_tree = Parser(Lexer(Logics.DQBF, dqbf_string)).getAST()
    
    skolem_functions = {}
    
    # Assume formulas in Skolem form
    uvars = parse_tree
    exvars = parse_tree.operands[0]
    
    univ = set()
    exis = set()
    exvar_scope = {}
    
    for uvar in uvars.variables:
        univ.add(convert_var(uvar.name))
    
    while exvars.isQuantifier():
        uvars = set()
        for uvar in exvars.scope:
            uvars.add(convert_var(uvar.name))
        for exvar in exvars.variables:
            exvar = convert_var(exvar.name)
            exis.add(exvar)
            exvar_scope[exvar] = uvars
        
        exvars = exvars.operands[0]
    
    #print(univ, exis, exvar_scope)
    
    visitor = PyEDAVisitor()
    exvars.accept(visitor)
    formula = visitor.formula
    
    print('cnf(rule_true,axiom, p(1)).')
    print('cnf(rule_false,axiom, ~p(0)).')
    
    if CNF_OUT:
        formula = formula.tseitin(auxvarname='aux')
    
        for i, f in enumerate(formula.args):
            literals = []
            for g in f.args:
                literal = ''
                if g.uniqid < 0:
                    literal += '~'
                    g = ~g
                if g in univ:
                    literal += 'p({})'.format(var2str(g, universal=True))
                else:
                    if g.name == 'aux':
                        depends = ['{}'.format(var2str(v, universal=True)) for v in univ]
                    else:
                        depends = ['{}'.format(var2str(v, universal=True)) for v in exvar_scope[g]]
                    literal += '{}({})'.format(var2str(g), ','.join(depends))
                literals.append(literal)
        
            print('cnf(clause{},plain,({})).'.format(i, ' | '.join(literals)))
    
    else:
        print('fof(goal,plain,{}).'.format(transform_formula_to_fof(formula)))
