# coding=utf-8

__author__ = "Leander Tentrup <tentrup@cs.uni-saarland.de>"
__copyright__ = "Copyright 2014, Saarland University"

from parse import *
import copy
from operator import mul
import subprocess, tempfile, os
from transformations import Negator, Flattening

def deps_compare(var1, var2, deps):
    if deps[var1] < deps[var2]: # proper subset?
        return -1
    elif deps[var1] == deps[var2]:
        return 0
    else:
        if var1.startswith('y') and not var2.startswith('y'):
            return -1
        elif not var1.startswith('y') and var2.startswith('y'):
            return 1
    return 0

def deps_exapand(deps):
    deps = copy.copy(deps)
    for var in deps:
        new_dep = True
        processed = []
        while new_dep:
            join = []
            new_dep = False
            for dep in deps[var]:
                if dep in deps and not dep in processed:
                    join.append(dep)
                    processed.append(dep)
            
            if join:
                new_dep = True
                for j in join:
                    deps[var] = deps[var].union(deps[j])
    
    return deps

'''
  Builds a QBF formula from a DQBF formula with bounded paths.
  Uses the bool2cnf tool (http://www-ise4.ist.osaka-u.ac.jp/~t-tutiya/sources/bool2cnf/)
  to transform the boolean formula in cnf.
'''
class DQBFVisitor(object):
    
    def __init__(self, num_paths, name_mapping=None):
        self.formula = None
        self.num_paths = num_paths
        self.operator_mapping = {
            TokenType.AND : '&',
            TokenType.OR : '|',
            TokenType.NOT : '!',
            TokenType.IMPL : '->',
            TokenType.EQUIV : '='
        }
        self.body = None
        self.current_path = -1
        self.consistency = []
        self.dqbf_analyzer = None

        if name_mapping:
            self.name_mapping = name_mapping
        else:
            self.name_mapping = lambda var, offset : var + str(offset)
        
        self.verify_herbrand = False
    
    def visitQuantifier(self, q):
        
        if not self.verify_herbrand:
            visitor = HerbrandVisitor()
            q.accept(visitor)
            self.verify_herbrand = True
            
            self.dqbf_analyzer = DQBFAnalyzer()
            q.accept(self.dqbf_analyzer)
        
        matrix = []
        if not q.operands[0].isQuantifier():
            for i in range(self.num_paths):
                self.current_path = i
                q.operands[0].accept(self) # result in self.formula
                matrix.append(self.formula)
        else:
            q.operands[0].accept(self) # result in self.formula
        
        if matrix:
            analyzer = self.dqbf_analyzer
            #print(analyzer.evariables)
            #print(analyzer.uvariables)
            #print(analyzer.dependencies)
            aux_vars = set(analyzer.evariables)
            for var in analyzer.pure_dependencies:
                for v in analyzer.pure_dependencies[var]:
                    aux_vars.discard(v)
            
            for var in aux_vars:
                if var in analyzer.dependencies:
                    del analyzer.dependencies[var]
            
            for var in analyzer.dependencies:
                analyzer.dependencies[var] -= aux_vars
            
            contained = analyzer.dependencies.keys()
            for var in contained:
                if not analyzer.dependencies[var]:
                    del analyzer.dependencies[var]
            
            variables = analyzer.evariables + analyzer.uvariables
            qorder = []
            deps = copy.copy(analyzer.dependencies)
            deps_expanded =  deps_exapand(deps)
            
            while analyzer.dependencies:
                i = 0
                keys = analyzer.dependencies.keys()
                keys.sort(cmp=lambda x, y: deps_compare(x, y, deps_expanded))
                    
                while i < len(keys):
                    
                    candidate = keys[i]
                    for var in analyzer.dependencies[candidate]:
                        if var in analyzer.dependencies:
                            candidate = None
                            break
                    
                    if candidate:
                        qorder.append(candidate)
                        del analyzer.dependencies[candidate]
                    
                    i += 1
            
            #print qorder
            # build consistency conditions:
            for var in qorder:
                if var in analyzer.uvariables:
                    for i in range(self.num_paths):
                        for j in range(i+1, self.num_paths):
                            disjunct = []
                            disjunct.append(BinOp(self.operator_mapping[TokenType.EQUIV],
                                                self.name_mapping(var, i),
                                                self.name_mapping(var, j))
                                            )
                            for dep_var in deps[var]:
                                disjunct.append(UnaryOp(self.operator_mapping[TokenType.NOT],
                                    BinOp(self.operator_mapping[TokenType.EQUIV],
                                        self.name_mapping(dep_var, i),
                                        self.name_mapping(dep_var, j))
                                    ))
                            self.consistency.append(BinOp(self.operator_mapping[TokenType.OR], operands=disjunct))
            
            if self.consistency:
                self.formula = BinOp(self.operator_mapping[TokenType.IMPL],
                        BinOp(self.operator_mapping[TokenType.AND], operands=self.consistency),
                        BinOp(self.operator_mapping[TokenType.OR], operands=matrix)
                    )
            else:
                self.formula = BinOp(self.operator_mapping[TokenType.OR], operands=matrix)
            self.body = self.formula
            
            #qorder = analyzer.evariables + analyzer.uvariables
            
            top_quantifier = None
            quantifier = None
            last_quantifier = None
            first_quantifier = None
            for var in qorder:
                #print(var)
                if var in analyzer.uvariables:
                    copies = [ self.name_mapping(var, i) for i in range(self.num_paths) ]
                    quantifier = Quantifier(TokenType.A, copies, None)
                    variables.remove(var)
                elif var in analyzer.evariables:
                    copies = [ self.name_mapping(var, i) for i in range(self.num_paths) ]
                    quantifier = Quantifier(TokenType.E, copies, None)
                    variables.remove(var)
                else:
                    assert False
                    
                if var in deps:
                    copies = []
                    #print(deps[var])
                    for v in deps[var]:
                        if v in variables:
                            copies += [ self.name_mapping(v, i) for i in range(self.num_paths) ]
                            variables.remove(v)
                    
                    if copies:
                        quantifier = Quantifier(TokenType.E, copies, quantifier)                
                        if not top_quantifier:
                            top_quantifier = quantifier
                
                if not top_quantifier:
                    top_quantifier = quantifier
                
                if last_quantifier:
                    last_quantifier.operands[0] = quantifier
                
                if quantifier.operands[0]:
                    last_quantifier = quantifier.operands[0]
                else:
                    last_quantifier = quantifier
            
            if not first_quantifier:
                first_quantifier = last_quantifier
            last_quantifier.operands[0] = self.formula
            self.formula = top_quantifier
            
            if variables:
                first_quantifier.operands[0] = Quantifier(TokenType.E, [self.name_mapping(v, i) for v in variables for i in range(self.num_paths)], self.body)
        
    def visitIdent(self, ident):
        assert False
        
    def visitProposition(self, prop):
        self.formula = copy.copy(prop)
        self.formula.name = self.name_mapping(prop.name, self.current_path)
        self.formula.ref = None
    
    def visitUnaryOp(self, unop):
        if unop.optype == TokenType.NOT:
            unop.operands[0].accept(self)
            self.formula = UnaryOp(self.operator_mapping[TokenType.NOT], self.formula)
        else:
            assert False


    def visitBinaryOp(self, binop):
        equiv = [TokenType.AND, TokenType.OR, TokenType.IMPL, TokenType.EQUIV]
        if binop.optype in equiv:
            operands = []
            
            assert not (binop.optype == TokenType.IMPL or binop.optype == TokenType.EQUIV) or (len(binop.operands) == 2)
            
            for o in binop.operands:
                o.accept(self)
                operands.append(self.formula)
            
            self.formula = BinOp(self.operator_mapping[binop.optype], operands=operands)
                
        elif binop.optype == TokenType.XOR:
            assert (len(binop.operands) == 2)
            binop.operands[0].accept(self)
            lhs = self.formula
            binop.operands[1].accept(self)
            rhs = self.formula
            
            self.formula = BinOp(self.operator_mapping[TokenType.AND],
                                BinOp(self.operator_mapping[TokenType.OR], lhs, rhs),
                                UnaryOp(self.operator_mapping[TokenType.NOT],
                                    BinOp(self.operator_mapping[TokenType.AND], lhs, rhs)
                                )
                            );

        else:
            assert False

    
    def __str__(self):
        return str(self.formula)

'''
    Verifies that the input DQBF is in Skolem/Herbrand form and transforms it to Herbrand form if needed.
'''
class HerbrandVisitor(object):
    
    def __init__(self):
        self.isSkolem = False
        self.isHerbrand = False
        self.prefix = True
        pass
    
    def visitQuantifier(self, q):
        
        if not self.isSkolem and not self.isHerbrand:
            # determine with the first quantifier if the formula is in skolem or herbrand form
            if q.qtype == TokenType.E and not q.scope:
                self.isHerbrand = True
            elif q.qtype == TokenType.A and not q.scope:
                self.isSkolem = True
            else:
                assert False
        
        if self.isSkolem:
            if self.prefix and q.qtype == TokenType.A:
                assert not q.scope
                q.qtype = TokenType.E
            elif q.qtype == TokenType.E:
                self.prefix = False
                q.qtype = TokenType.A
            else:
                assert False
                
            if q.operands[0].isQuantifier():
                q.operands[0].accept(self)
            else:
                negator = Negator()
                q.operands[0].accept(negator)
                flattener = Flattening()
                negator.formula.accept(flattener)
                q.operands[0] = flattener.formula
        
    def visitIdent(self, ident):
        assert False
        
    def visitProposition(self, prop):
        assert False
    
    def visitUnaryOp(self, unop):
        assert False

    def visitBinaryOp(self, binop):
        assert False


class DQBFAnalyzer(object):
    
    def __init__(self):
        self.dependencies = {}
        self.pure_dependencies = {}
        self.uvariables = []
        self.evariables = []
        self.matrix = None
        self.dependency = set()
    
    def visitQuantifier(self, q):
        
        if q.qtype == TokenType.E:
            assert q.scope is None
            self.evariables += [p.name for p in q.variables]
        else:
            assert q.scope is not None
            uvars = [p.name for p in q.variables]
            scope = [p.name for p in q.scope]
            self.uvariables += uvars
            for uvar in uvars:
                self.dependencies[uvar] = set(scope)
                self.pure_dependencies[uvar] = set(scope)
        
        if not q.operands[0].isQuantifier():
            self.matrix = q.operands[0]
        
        q.operands[0].accept(self)
        
    def visitIdent(self, ident):
        assert False
        
    def visitProposition(self, prop):
        self.dependency.add(prop.name)
    
    def visitUnaryOp(self, unop):
        unop.operands[0].accept(self)

    def visitBinaryOp(self, binop):
        if binop == self.matrix:
            for o in binop.operands:
                self.checkDependency(o)
        else:
            for o in binop.operands:
                o.accept(self)
    
    def checkDependency(self, op):
        try:
            if op.optype == TokenType.EQUIV and op.operands[0].name:
                self.dependency = set()
                var = op.operands[0].name
                op.operands[1].accept(self)
                assert var not in self.dependencies
                self.dependencies[var] = self.dependency
                
        except AttributeError:
            pass
        