#!/usr/bin/env python                                                    
# coding=utf-8

"""Expansion based DQBF solver using the CUDD BDD library."""

__author__ = "Leander Tentrup <tentrup@cs.uni-saarland.de>"
__copyright__ = "Copyright 2014, Saarland University"

import sys, os, re, copy

from lex import *
from parse import *

import pycudd

CUDD_REORDER_SAME = 0
CUDD_REORDER_RANDOM = 2
CUDD_REORDER_SIFT = 4
CUDD_REORDER_SIFT_CONVERGE = 5
CUDD_REORDER_LAZY_SIFT = 20

def printUsage():
    print('Usage: %s ["DQBF string" | -F file_name | stdin]' % sys.argv[0])
    exit(1)


'''
  Transforms boolean formula to CUDD BDD format
'''
class CUDDVisitor(object):
    
    def __init__(self, bdd_solver):
        self.formula = None
        self.bdd_solver = bdd_solver
    
    def visitQuantifier(self, q):
        assert False
         
    def visitIdent(self, ident):
        assert False
         
    def visitProposition(self, prop):
        self.formula = self.bdd_solver.getVariable(prop.name)
     
    def visitUnaryOp(self, unop):
        unop.operands[0].accept(self)
        assert unop.optype == TokenType.NOT
        self.formula = ~self.formula
     
    def visitBinaryOp(self, binop):
        assert len(binop.operands) == 2
        
        binop.operands[0].accept(self)
        lhs = self.formula
        
        binop.operands[1].accept(self)
        
        if binop.optype == TokenType.AND:
            self.formula &= lhs
        elif binop.optype == TokenType.OR:
            self.formula |= lhs
        elif binop.optype == TokenType.XOR:
            self.formula = lhs ^ self.formula
        elif binop.optype == TokenType.IMPL:
            self.formula = (~lhs) | self.formula
        elif binop.optype == TokenType.EQUIV:
            self.formula = lhs.Xnor(self.formula)
        else:
            assert False


class BDDSolver:
    def __init__(self, parse_tree):
        # pyCUDD manager
        self.mgr = pycudd.DdManager()
        self.mgr.SetDefault()
    
        # Optimizations
        self.mgr.AutodynEnable(CUDD_REORDER_LAZY_SIFT)
        
        # transform parse tree into BDD
        self.variables = {} # name => variable index
        self.variable_count = 0
        
        # Assume formulas in Skolem form
        uvars = parse_tree
        exvars = parse_tree.operands[0]
    
        self.universals = set()
        self.existentials = set()
        self.existential_scope = {}
    
        for uvar in uvars.variables:
            self.universals.add(self.getVariable(uvar.name, returnid=True))
    
        while exvars.isQuantifier():
            dependent = set()
            for uvar in exvars.scope:
                dependent.add(self.getVariable(uvar.name, returnid=True))
            for exvar in exvars.variables:
                exvar = self.getVariable(exvar.name, returnid=True)
                self.existentials.add(exvar)
                self.existential_scope[exvar] = dependent
        
            exvars = exvars.operands[0]
    
        #print(self.universals, self.existentials, self.existential_scope)
        
        visitor = CUDDVisitor(self)
        exvars.accept(visitor)
        self.formula = visitor.formula
    
    def getVariable(self, name, returnid=False):
        if name in self.variables:
            return self.mgr.ReadVars(self.variables[name]) if not returnid else self.variables[name]
        else:
            bdd_var = self.mgr.NewVar()
            self.variables[name] = self.variable_count
            self.variable_count += 1
            
            return bdd_var if not returnid else self.variable_count - 1
    
    def newVariable(self):
        self.mgr.NewVar()
        self.variable_count += 1
        
        return self.variable_count - 1

    def isSatisfiable(self):
        while self.universals:
            fully_informed_exvar = self.dependsOn(self.universals)
    
            if fully_informed_exvar:
                # eliminate existential
                cube = self.getCubeFromIndices(fully_informed_exvar)
                self.formula = self.formula.ExistAbstract(cube)
                
                for exvar in fully_informed_exvar:
                    del self.existential_scope[exvar]
    
                self.existentials -= fully_informed_exvar
    
            independent_universals = self.independentUniversals()
            if independent_universals:
                # eliminate universals
                cube = self.getCubeFromIndices(independent_universals)
                self.formula = self.formula.UnivAbstract(cube)
    
                self.universals -= independent_universals
            
            if not self.universals:
                break
            
            uvar = self.minimalDependent(self.universals)
            # eliminate universal
            self.universals.remove(uvar)
            permutation = range(self.variable_count)
            eval_one = self.mgr.ReadVars(uvar)
            eval_zero = ~self.mgr.ReadVars(uvar)
            for exvar in self.dependsOn(uvar):
                # copy existential
                self.existential_scope[exvar].remove(uvar)
        
                copy_exvar = self.newVariable()
                permutation[exvar] = copy_exvar
                
                assert copy_exvar not in self.existentials
                self.existentials.add(copy_exvar)
                self.existential_scope[copy_exvar] = copy.copy(self.existential_scope[exvar])
            
            permutation,_ = self.IntArrayFromList(permutation)
            
            self.formula = self.formula.Restrict(eval_one) & self.formula.Restrict(eval_zero).Permute(permutation)
    
        for exvar in self.existential_scope:
            assert not self.existential_scope[exvar]
    
        return bool(self.formula)

    def dependsOn(self, uvar):
        ''' returns the existential variable which depends on uvar if
            uvar is a single universal variable, or all existential 
            variables with dependencies uvar if uvar is a set of variables'''
        exvars = set()
        for exvar in self.existential_scope:
            if uvar in self.existential_scope[exvar] or uvar == self.existential_scope[exvar]:
                exvars.add(exvar)
        return exvars
        
    def minimalDependent(self, univ):
        ''' get a universal variable which is the fewest contained in the support sets '''
        assert univ
        minuvar = None
        contained = float("Inf")
        for uvar in univ:
            dependent = len(self.dependsOn(uvar))
            if dependent < contained:
                minuvar = uvar
                contained = dependent
    
        return minuvar
    
    def independentUniversals(self):
        universals = copy.copy(self.universals)
        for exvar in self.existential_scope:
            universals -= self.existential_scope[exvar]
        return universals
    
    def getCubeFromIndices(self, indices):
        indice_list = []
        for i in indices:
            indice_list.append(i)
            
        # build index cube
        array, size = self.IntArrayFromList(indice_list)
        return self.mgr.IndicesToCube(array, size)
        
    def IntArrayFromList(self, index_list):
        size = len(index_list)
        array = pycudd.IntArray(size)
        for i in xrange(size):
            array[i] = index_list[i]
        
        return array, size

if __name__ == "__main__":
    
    #sys.setrecursionlimit(100000)
    
    process_stdin = False
    if not sys.stdin.isatty():
        process_stdin = True
        dqbf_string = sys.stdin.read()
        
        if len(sys.argv) != 1:
            printUsage()
    
    elif len(sys.argv) == 2:
        dqbf_string = sys.argv[1]
    elif len(sys.argv) == 3 and sys.argv[1] == '-F':
        file_handle = open(sys.argv[2], 'r')
        dqbf_string = file_handle.read()
    else:
        printUsage()
    
    parse_tree = Parser(Lexer(Logics.DQBF, dqbf_string)).getAST()
    
    solver = BDDSolver(parse_tree)
    
    if solver.isSatisfiable():
        print('SAT')
        exit(10)
    else:
        print('UNSAT')
        exit(20)

            