# coding=utf-8

"""Lexer for various logics."""

__author__ = "Leander Tentrup <tentrup@cs.uni-saarland.de>"
__copyright__ = "Copyright 2014, Saarland University"

class Logics:
    ECL  = 'ECL'
    QPTL = 'QPTL'
    QBF  = 'QBF'
    DQBF = 'DQBF'
    CTL  = 'CTL'

class TokenType:
  # Quantifiers
  A        = 'A'
  E        = 'E'
  
  COLON    = ':'

  # Parens
  LPAREN   = '('
  RPAREN   = ')'
  LBRACE   = '{'
  RBRACE   = '}'

  # Operators
  AND      = '/\\'
  OR       = '\\/'
  NOT      = '~'
  UNTIL    = 'U'
  WUNTIL   = 'W'
  RELEASE  = 'R'
  NEXT     = '()'
  GLOBALLY = '[]'
  FINALLY  = '<>'
  IMPL     = '-->'
  EQUIV    = '<-->'
  XOR      = '^'
  
  # CTL
  ALWAYSGLOBALLY = 'AG'

  # Variables and Literals
  IDENT    = 'IDENT'
  TRUE     = 'true'
  FALSE    = 'false'

  EOF      = '\n'
  
  LINECOMMENT = '#'
  
  CHARS = {
    Logics.QBF  : [A,E,COLON,LPAREN,RPAREN,AND,OR,NOT,IMPL,EQUIV],
    Logics.DQBF : [A,E,COLON,LPAREN,RPAREN,LBRACE,RBRACE,AND,OR,NOT,IMPL,EQUIV,XOR],
    Logics.QPTL : [A,E,COLON,LPAREN,RPAREN,AND,OR,NOT,UNTIL,WUNTIL,RELEASE,NEXT,GLOBALLY,FINALLY,IMPL,EQUIV],
    Logics.ECL  : [A,E,COLON,LPAREN,RPAREN,LBRACE,RBRACE,AND,OR,NOT,UNTIL,WUNTIL,RELEASE,NEXT,GLOBALLY,FINALLY,IMPL,EQUIV],
    Logics.CTL  : [LPAREN,RPAREN,AND,OR,NOT,ALWAYSGLOBALLY,EQUIV,TRUE,FALSE]
  }
  
  QUANTIFIER  = [A,E]
  BINARYOP    = [AND,OR,UNTIL,WUNTIL,RELEASE,IMPL,EQUIV,XOR]
  UNARYOP     = [NOT,NEXT,GLOBALLY,FINALLY,ALWAYSGLOBALLY]
  
  ALIAS = {
      '&' : AND,
      '|' : OR,
      '!' : NOT,
      '-' : NOT,
      '->' : IMPL,
      '=' : EQUIV,
      '0' : FALSE,
      '1' : TRUE,
      '+' : OR,
      '*' : AND
  }
  
  def __init__(self, logic):
      if logic not in self.CHARS:
          raise Exception('unknown logic %(logic)s, possible logics are %(logics)s' % {'logic':logic, 'logics':', '.join(CHARS.keys())})

      self.chars = {}
      for char in self.CHARS[logic]:
          self.append_char(char)
          
          for alias in self.ALIAS:
              if self.ALIAS[alias] == char:
                  self.append_char(alias)

  def append_char(self, char):
      if len(char) not in self.chars:
          self.chars[len(char)] = [list(char)]
      else:
          self.chars[len(char)].append(list(char))

class Token:
  def __init__(self, ttype, value=None):
    self.ttype = ttype
    self.value = value

  def isType(self, ttype):
    return ttype == self.ttype

  def isQuantifier(self):
    return self.ttype in TokenType.QUANTIFIER

  def isBinaryOp(self):
    return self.ttype in TokenType.BINARYOP

  def isUnaryOp(self):
    return self.ttype in TokenType.UNARYOP

  def __str__(self):
    val = self.ttype
    if self.value:
      val += ' ' + self.value
    return val

class Lexer:
    def __init__(self, logic, string):
        self.logic = logic
        self.tokentype = TokenType(logic)
        self.charstream = list(string)
        self.charpointer = 0
        self.streamlength = len(self.charstream)
        self.tokenlength = []
        for i in reversed(sorted(self.tokentype.chars.keys())):
            self.tokenlength.append(i)

    def nextToken(self):
        
        if self.charstream is None:
            return None

        elif self.charpointer >= self.streamlength:
            self.charstream = None
            return Token(TokenType.EOF)

        elif self.charstream[self.charpointer].isspace():
            self.charpointer += 1
            return self.nextToken()
        
        elif self.charstream[self.charpointer] == TokenType.LINECOMMENT:
            self.charpointer += 1
            while self.charstream[self.charpointer] != '\n':
                self.charpointer += 1
            return self.nextToken()
    
        else:
            for i in self.tokenlength:
                if self.charstream[self.charpointer:self.charpointer+i] in self.tokentype.chars[i]:
                    keyword = ''.join(self.charstream[self.charpointer:self.charpointer+i])
                    self.charpointer += i
                    
                    # Check for alias
                    if keyword in TokenType.ALIAS:
                        keyword = TokenType.ALIAS[keyword]
                    
                    return Token(keyword)
    
            if self.charstream[self.charpointer].isalpha():
                string = ''
                while self.charpointer < self.streamlength and (self.charstream[self.charpointer].isalpha() or self.charstream[self.charpointer].isdigit() or self.charstream[self.charpointer] == '_'):
                    string += self.charstream[self.charpointer]
                    self.charpointer += 1
      
                if string == TokenType.TRUE or (string in TokenType.ALIAS and TokenType.ALIAS[string] == TokenType.TRUE):
                    return Token(TokenType.TRUE)
                elif string == TokenType.FALSE or (string in TokenType.ALIAS and TokenType.ALIAS[string] == TokenType.FALSE):
                    return Token(TokenType.FALSE)
                else:
                    return Token(TokenType.IDENT, string)
            
            elif self.charstream[self.charpointer] == "'":
                string = ''
                self.charpointer += 1
                while self.charstream and self.charstream[self.charpointer] != "'":
                    string += self.charstream[self.charpointer]
                    self.charpointer += 1
                
                self.charpointer += 1
                return Token(TokenType.IDENT, string)

            else:
                print(self.charstream)
                raise Exception('unknown Token ' + self.charstream[self.charpointer] + ' at position ' + str(self.charpointer))
