# coding=utf-8

"""Parser for various logics."""

__author__ = "Leander Tentrup <tentrup@cs.uni-saarland.de>"
__copyright__ = "Copyright 2014, Saarland University"

from lex import *

class Parser:
  def __init__(self, lexer):
    self.lexer = lexer
    self.current = self.lexer.nextToken()
    self.next = self.lexer.nextToken()
    self.proposition_lookup = {}

  def getAST(self):
    ast = self.parse()

    if not self.current.isType(TokenType.EOF):
      raise Exception("Expected end of formula, found %(token)s" % {'token': self.current})

    return ast

  def parse(self):
    if self.current.isQuantifier():
        return self.parseQuantifier()
    else:
        return self.parseExpression(0)

  def parseIdent(self):
    assert self.current.isType(TokenType.IDENT)

    value = self.current.value
    self.nextToken()
    return Identifier(value)

  def parseQuantifier(self):
    assert self.current.isQuantifier()
    qtype = self.current.ttype
    self.nextToken()
    
    scope = []
    if (self.lexer.logic in Logics.ECL) or (self.current.isType(TokenType.LBRACE) and self.lexer.logic in Logics.DQBF):
        self.nextToken(current=TokenType.LBRACE)
        while (self.current.isType(TokenType.IDENT)):
            ident = self.parseIdent()
            if ident.name in self.proposition_lookup:
                ident = self.proposition_lookup[ident.name]
            scope.append(ident)
        self.nextToken(current=TokenType.RBRACE)
    
    variables = []
    while (self.current.isType(TokenType.IDENT)):
        ident = self.parseIdent()
        if ident.name in self.proposition_lookup:
            raise Exception('Variable %s can only be introduced once!' % ident.name)
        self.proposition_lookup[ident.name] = ident
        variables.append(ident)
    
    if len(variables) == 0:
        raise Exception("Expected variable after quantification")
    
    self.nextToken(current = TokenType.COLON)
    
    if self.lexer.logic == Logics.ECL or (scope and self.lexer.logic == Logics.DQBF):
        return Quantifier(qtype, variables, self.parse(), scope)
    else:
        return Quantifier(qtype, variables, self.parse())

  # Unary Operators > Temporal Binary Operators > Conjunction > Disjunction > Implication > Equivalence > Quantification
  # Operator:    <-->  -->  \/  /\  U|W|R
  # Precedence:   0     1   2   3     4
  def parseExpression(self, min_precedence):
      lhs = self.parseUnaryExpression()
      
      while self.current.isBinaryOp() and self.getPrecedence(self.current.ttype) >= min_precedence:
          op = self.current.ttype
          self.nextToken()
          rhs = self.parseExpression(self.getPrecedence(op) + 1)
          lhs = BinOp(op, lhs, rhs)
      
      return lhs

  def parseUnaryExpression(self):
      if self.current.isUnaryOp():
          op = self.current.ttype
          self.nextToken()
          expr = self.parseUnaryExpression()
          return UnaryOp(op, expr)
      else:
          return self.parsePrimaryExpression()

  def parsePrimaryExpression(self):
      if self.current.isType(TokenType.IDENT):
          name = self.current.value
          self.nextToken()
          ref = None
          if name in self.proposition_lookup:
              ref = self.proposition_lookup[name]
          return Proposition(name, ref)
      elif self.current.isType(TokenType.FALSE):
          self.nextToken()
          return Literal(0)
      elif self.current.isType(TokenType.TRUE):
          self.nextToken()
          return Literal(1)
      elif self.current.isType(TokenType.LPAREN):
          self.nextToken()
          expr = self.parse()
          self.nextToken(current = TokenType.RPAREN)
          return expr
      else:
          raise Exception('Expected identifier or parenthesis, found %s' % str(self.current))
          
  def getPrecedence(self, op):
      precedences = {
        TokenType.EQUIV   : 0,
        TokenType.XOR     : 0,
        TokenType.IMPL    : 1,
        TokenType.OR      : 2,
        TokenType.AND     : 3,
        TokenType.UNTIL   : 4,
        TokenType.WUNTIL  : 4,
        TokenType.RELEASE : 4
      }
      return precedences[op]

  def nextToken(self, expected = None, current = None):
    if expected is not None and not self.next.isType(expected):
      raise Exception("Expected %(expected)s, found %(found)s", {'expected':expected, 'found':str(self.next)});
    if current is not None and not self.current.isType(current):
      raise Exception("Expected %(expected)s, found %(found)s", {'expected':current, 'found':str(self.current)});

    self.current = self.next
    self.next = self.lexer.nextToken()
    

'''
'' Base class for AST, has a list of operands
'''
class AST(object):
  def __init__(self, operands):
    self.operands = operands

  def accept(self, visitor):
      assert False
      #visitor.visit(self)
      #for operand in self.operands:
      #    operand.accept(visitor)

  def isQuantifier(self):
      return False
  def isLiteral(self):
      return False

'''
'' Class for all quantifier constructs (A x1 x2 … :, E x1 x2 … :)
'''
class Quantifier(AST):
  def __init__(self, qtype, variables, formula, scope=None):
    AST.__init__(self, [formula])
    self.qtype = qtype
    self.variables = variables
    self.scope = scope
  
  def __str__(self):
      rep = self.qtype
      if self.scope:
          rep += '{%s}' % ' '.join([str(v) for v in self.scope])
      rep += ' '
      rep += ' '.join([str(v) for v in self.variables])
      rep += ' : ' + str(self.operands[0])
      return rep

  def accept(self, visitor):
      visitor.visitQuantifier(self)

  def isQuantifier(self):
      return True

class Identifier(AST):
  def __init__(self, name):
    AST.__init__(self, [])
    self.name = name
    
  def __str__(self):
      return self.name

  def accept(self, visitor):
      visitor.visitIdentifier(self)
      
class Proposition(AST):
  def __init__(self, name, ref=None):
    AST.__init__(self, [])
    self.name = name
    self.ref = ref
    
  def __str__(self):
      return self.name

  def accept(self, visitor):
      visitor.visitProposition(self)
      
      
class Literal(AST):
  def __init__(self, val):
    AST.__init__(self, [])
    self.val = val
    
  def __str__(self):
      return '%d' % self.val

  def accept(self, visitor):
      visitor.visitProposition(self)
      
  def isLiteral(self):
      return True

class UnaryOp(AST):
  def __init__(self, optype, formula):
    AST.__init__(self, [formula])
    self.optype = optype

  def __str__(self):
      return '(' + self.optype + ' ' + str(self.operands[0]) + ')'

  def accept(self, visitor):
      visitor.visitUnaryOp(self)

class BinOp(AST):
  def __init__(self, optype, left=None, right=None, operands=None):
    if left and right:
      AST.__init__(self, [left, right])
    else:
      assert operands
      AST.__init__(self, operands)
    self.optype = optype

  def __str__(self):
      #return '(' + str(self.operands[0]) + ' ' + self.optype + ' ' + str(self.operands[1]) + ')'
      #output = '(' + str(self.operands[0]) + ' ' + self.optype + ' ' + str(self.operands[1]) + ')'
      #for i in xrange(2, len(self.operands)):
      #    output = '(' + output + ' ' + self.optype + ' ' + str(self.operands[i]) + ')'
      #return output
      return '(' + (' ' + self.optype + ' ').join([str(o) for o in self.operands]) + ')'

  def accept(self, visitor):
      visitor.visitBinaryOp(self)
      