#!/usr/bin/env python
# coding=utf-8

"""Generates DQBF formulas from PEC instances."""

__author__ = "Leander Tentrup <tentrup@cs.uni-saarland.de>"
__copyright__ = "Copyright 2014, Saarland University"

"""
Description:
- Generate PEC examples using a variable number of black boxes.
- The black boxes are selected randomly from all possible gates.
- Errors are introduced by randomly exchanging gates.
"""

import os, sys, random, copy, argparse

num_y_vars = 0
yvars = {}
num_inputs = 0
num_aux_vars = 0
bbs_gates_processed = {} # gate -> yvar
count_gates_processed = {} # gate -> var
bits = -1
use_circuit_aux_vars = True
black_box_size = 1

EXAMPLE_CIRCUIT    = 'example_circuit' # From ``Checking Equivalence for Partial Implementations'' by C. Scholl and B. Becker
FULL_ADDER         = 'full_adder'
ADDER_3_BIT        = 'adder_3_bit'
ADDER_N_BIT        = 'adder_n_bit'
# From ``Digital Design, a practical approach'' by W. J. Dally, R. C. Harting, p. 167
BIT_CELL_ARBITER   = 'bit_cell_arbiter' 
LOOK_AHEAD_ARBITER = 'look_ahead_arbiter'
# From ``Digital Design, a practical approach'' by W. J. Dally, R. C. Harting, p. 224
MULTIPLIER         = 'multiplier'

MULTIPLEXER        = 'multiplexer'

circuits = (EXAMPLE_CIRCUIT, FULL_ADDER, ADDER_3_BIT, ADDER_N_BIT, BIT_CELL_ARBITER, LOOK_AHEAD_ARBITER, MULTIPLIER, MULTIPLEXER)

class Gate(object):
    AND   = 'AND'
    NAND  = 'NAND'
    OR    = 'OR'
    NOR   = 'NOR'
    XOR   = 'XOR'
    XNOR  = 'XNOR'
    NOT   = 'NOT'
    GATES = (AND, NAND, OR, NOR, XOR, XNOR)
    
    cache = {} # (type, operands) => gate
    
    def __init__(self, gate_type, operands):
        assert gate_type == Gate.NOT or len(operands) == 2
        assert not gate_type == Gate.NOT or len(operands) == 1
        self.gate_type = gate_type
        self.operands = operands
        self.count = 0
        self.aux_var = False
    
    def appearsInCircuit(self):
        self.count += 1
    
    def isGate(self):
        return True
        
    def isInput(self):
        return False
    
    def __str__(self):
        return '{} {} {}'.format(self.__hash__(), self.gate_type, self.count)
    
    def toDot(self):
        dot = set(['{}[shape=box][label="{}"];'.format(self.__hash__(), self.gate_type)])
        for o in self.operands:
            dot.add('{} -> {};'.format(self.__hash__(), o.__hash__()))
            dot.update(o.toDot())
        return dot

AND_OP  = '({} & {})'
NAND_OP = '!({} & {})'
OR_OP   = '({} | {})'
NOR_OP  = '!({} | {})'
XOR_OP  = '({} ^ {})'
XNOR_OP = '({} <--> {})'

bool_ops = [AND_OP, NAND_OP, OR_OP, NOR_OP, XOR_OP, XNOR_OP]

gate_translation = {
    Gate.AND  : AND_OP,
    Gate.NAND : NAND_OP,
    Gate.OR   : OR_OP,
    Gate.NOR  : NOR_OP,
    Gate.XOR  : XOR_OP,
    Gate.XNOR : XNOR_OP
}

def And(left, right):
    if not (Gate.AND, left, right) in Gate.cache:
        gate = Gate(Gate.AND, [left, right])
        Gate.cache[(Gate.AND, left, right)] = gate
    return Gate.cache[(Gate.AND, left, right)]
        
def Nand(left, right):
    if not (Gate.NAND, left, right) in Gate.cache:
        gate = Gate(Gate.NAND, [left, right])
        Gate.cache[(Gate.NAND, left, right)] = gate
    return Gate.cache[(Gate.NAND, left, right)]

def Or(left, right):
    if not (Gate.OR, left, right) in Gate.cache:
        gate = Gate(Gate.OR, [left, right])
        Gate.cache[(Gate.OR, left, right)] = gate
    return Gate.cache[(Gate.OR, left, right)]

def Nor(left, right):
    if not (Gate.NOR, left, right) in Gate.cache:
        gate = Gate(Gate.NOR, [left, right])
        Gate.cache[(Gate.NOR, left, right)] = gate
    return Gate.cache[(Gate.NOR, left, right)]

def Xor(left, right):
    if not (Gate.XOR, left, right) in Gate.cache:
        gate = Gate(Gate.XOR, [left, right])
        Gate.cache[(Gate.XOR, left, right)] = gate
    return Gate.cache[(Gate.XOR, left, right)]

def Xnor(left, right):
    if not (Gate.XNOR, left, right) in Gate.cache:
        gate = Gate(Gate.XNOR, [left, right])
        Gate.cache[(Gate.XNOR, left, right)] = gate
    return Gate.cache[(Gate.XNOR, left, right)]

def Not(operand):
    if not (Gate.NOT, operand) in Gate.cache:
        gate = Gate(Gate.NOT, [operand])
        Gate.cache[(Gate.NOT, operand)] = gate
    return Gate.cache[(Gate.NOT, operand)]

class Input(object):
    def __init__(self, name):
        assert name not in Input.input_cache and "Use Input.get to create inputs"
        self.name = name

    def isGate(self):
        return False
        
    def isInput(self):
        return True
    
    def toDot(self):
        return set(['{}[shape=box][label="{}"];'.format(self.__hash__(), self.name)])
    
    input_cache = {} # str => Input
    
    @staticmethod
    def get(name):
        if name not in Input.input_cache:
            var = Input(name)
            Input.input_cache[name] = var
        return Input.input_cache[name]

class Circuit(object):
    def __init__(self, outputs):
        self.outputs = outputs # [Gate/Input]
        self.gates = []
    
    def addOutput(self, outputs):
        assert not self.gates
        try:
            self.outputs.extend(outputs)
        except TypeError:
            self.outputs.append(outputs)
    
    def combine(self, circuit):
        assert not self.gates
        assert not circuit.gates
        
        self.outputs.extend(circuit.outputs)
    
    def pop(self):
        assert not self.gates
        return self.outputs.pop()
    
    def getGates(self):
        ''' Lists all gates that appear in the circuit. Call only once! '''
        assert not self.gates
        for obj in self.outputs:
            self._getGates(obj)
        for obj in self.outputs:
            self.determineAuxVars(obj, 1)
        return self.gates
    
    def _getGates(self, obj):
        if obj.isGate():
            obj.appearsInCircuit()
            if obj not in self.gates:
                self.gates.append(obj)
            for o in obj.operands:
                self._getGates(o)
    
    def determineAuxVars(self, obj, last_count):
        if obj.isGate():
            if obj.count > last_count:
                obj.aux_var = True
                last_count = obj.count
            for o in obj.operands:
                self.determineAuxVars(o, last_count)

def halfAdder(i1, i2):
    and_i1_i2 = And(i1, i2)
    xor_i1_i2 = Xor(i1, i2)
    
    return Circuit([and_i1_i2, xor_i1_i2])

def fullAdder(i1, i2, cin):
    # s
    xor_i1_i2 = Xor(i1, i2)
    xor_1_cin = Xor(xor_i1_i2, cin)
    # cout
    and_1_cin = And(xor_i1_i2, cin)
    and_i1_i2 = And(i1, i2)
    or_2_3    = Or(and_1_cin, and_i1_i2)
    
    return Circuit([xor_1_cin, or_2_3])

def bitCell(req, priority):
    # grant
    grant = And(req, priority)
    # new priority
    new_prio = And(Not(req), priority)
    
    return Circuit([grant, new_prio])

def lookAheadArbiter(bits):
    in_vars = [Input.get('x{}'.format(i)) for i in xrange(1, bits+1)]
    output_bits = []
    gates_lookup = {}
    for i in xrange(0,bits):
        if i == 0:
            o = in_vars[0]
        else:
            first_round = True
            inputs = [in_vars[j] for j in xrange(0, i)]
            while len(inputs) != 1:
                inter_gates = []
                while len(inputs) >= 2:
                    lhs = inputs.pop(0)
                    rhs = inputs.pop(0)
                
                    if first_round and not inputs:
                        lhs = Not(lhs)
                    elif first_round and inputs:
                        lhs = Not(lhs)
                        rhs = Not(rhs)
                    
                    if not lhs in gates_lookup:
                        gates_lookup[lhs] = {}
                    if not rhs in gates_lookup[lhs]:
                        gate = And(lhs, rhs)
                        gates_lookup[lhs][rhs] = gate
                    
                    gate = gates_lookup[lhs][rhs]
                    inter_gates.append(gate)
                    
                inputs.extend(inter_gates)
                first_round = False
            o = inputs[0]
        assert o
        output_bits.append(o)
    
    return Circuit(output_bits)

def multiplierCircuit(num):
    in_vars = [Input.get('x{}'.format(i+1)) for i in xrange(2 * num)]
    in_false = Input.get('0')
    in_true = Input.get('1')
    gates = []
    matrix = [num*[None] for i in xrange(num)]
    last_carry = None
    for row in xrange(num):
        for col in xrange(num):
            input_row = num + row
            input_col = col
            if row == 0:
                and_gate = And(in_vars[input_col], in_vars[input_row])
                matrix[row][col] = and_gate
                gates.append(and_gate)
            else:
                and_gate = And(in_vars[input_col], in_vars[input_row])
                gates.append(and_gate)
                
                if col == num - 1:
                    if last_carry == None:
                        rhs = in_false
                    else:
                        rhs = last_carry
                else:
                    rhs = matrix[row - 1][col + 1]
                
                if col == 0:
                    cin = in_false
                else:
                    cin = cout
                
                adder = fullAdder(and_gate, rhs, cin)
                out, cout = adder.outputs
                
                if col == num - 1:
                    last_carry = cout
                
                matrix[row][col] = out
    
    output_bits = []
    for row in xrange(num):
        output_bits.append(matrix[row][0])
    for col in xrange(1,num):
        output_bits.append(matrix[num - 1][col])
    output_bits.append(last_carry) # add last carry
    
    assert len(output_bits) == 2*num
    
    return Circuit(output_bits)
    
def multiplexerCircuit(log_bits):
    num_inputs = 2**log_bits
    
    input_vars = [Input.get('x{}'.format(i+1)) for i in xrange(num_inputs)]
    controll_vars = [Input.get('x{}'.format(i+1)) for i in xrange(num_inputs,num_inputs+log_bits)]
    ncontroll_vars = [Not(cvar) for cvar in controll_vars]
    
    or_outs = []
    
    for i, ivar in enumerate(input_vars):
        binary = bin(i)[2:]
        binary = '0'*(log_bits - len(binary)) + binary
        and_out = []
        for j, b in enumerate(reversed(binary)):
            b = int(b)
            if b:
                and_out.append(controll_vars[j])
            else:
                and_out.append(ncontroll_vars[j])
        and_out.append(ivar)
        
        # Build and gates
        while len(and_out) >= 2:
            lhs = and_out.pop(0)
            rhs = and_out.pop(0)
            
            and_out.append(And(lhs, rhs))
        
        assert len(and_out) == 1
        or_outs.extend(and_out)
    
    
    while len(or_outs) >= 2:
        lhs = or_outs.pop(0)
        rhs = or_outs.pop(0)
        
        or_outs.append(Or(lhs, rhs))
    
    return Circuit(or_outs)

def exampleCircuit():
    x1, x2, x3, x4, x5, x6, x7, x8 = [Input.get('x{}'.format(i+1)) for i in xrange(8)]
    xor_x2_x3 = Xor(x2, x3)
    and_x4_x5 = And(x4, x5)
    or_x6_x7  = Or(x6, x7)
    xor_1_2   = Xor(xor_x2_x3, and_x4_x5)
    and_2_nx8 = And(and_x4_x5, Not(x8))
    and_3_x8  = And(or_x6_x7, x8)
    or_x1_4   = Or(x1, xor_1_2)
    or_5_6    = Or(and_2_nx8, and_3_x8)
    
    return Circuit([or_x1_4, or_5_6])

def buildCircuit(circuit_type, num_bbs, error=False):
    global num_inputs, bits
    
    circuit = None
    
    if circuit_type == EXAMPLE_CIRCUIT:
        num_inputs = 8
        circuit = exampleCircuit()
    elif circuit_type == FULL_ADDER:
        num_inputs = 3
        x1, x2, x3 = [Input.get('x{}'.format(i+1)) for i in xrange(num_inputs)]
        circuit = fullAdder(x1, x2, x3)
    elif circuit_type == ADDER_3_BIT:
        num_inputs = 6
        x1, x2, x3, x4, x5, x6 = [Input.get('x{}'.format(i+1)) for i in xrange(num_inputs)]
        
        circuit = halfAdder(x1, x4)
        c = circuit.pop()
        circuit.combine(fullAdder(x2, x5, c))
        c = circuit.pop()
        circuit.combine(fullAdder(x3, x6, c))
        
    elif circuit_type == ADDER_N_BIT:
        assert bits > 2
        num_inputs = 2 * bits
        in_vars = [Input.get('x{}'.format(i+1)) for i in xrange(num_inputs)]
        circuit = None
        c = None
        for i in xrange(bits):
            if i == 0:
                circuit = halfAdder(in_vars[i], in_vars[bits + i])
                c = circuit.pop()
            else:
                circuit.combine(fullAdder(in_vars[i], in_vars[bits + i], c))
                c = circuit.pop()
        circuit.addOutput(c)
        
    elif circuit_type == BIT_CELL_ARBITER:
        if bits == -1:
            bits = 3
        num_inputs = bits + 1
        in_vars = [Input.get('x{}'.format(i+1)) for i in xrange(num_inputs)]
        circuit = Circuit([])
        
        prio = Input.get('1')
        for i in xrange(num_inputs):
            circuit.combine(bitCell(in_vars[i], prio))
            prio = circuit.pop()
        circuit.addOutput(prio)
        
    elif circuit_type == LOOK_AHEAD_ARBITER:
        if bits == -1:
            bits = 4
        num_inputs = bits
        circuit = lookAheadArbiter(bits)
        
    elif circuit_type == MULTIPLIER:
        if bits == -1:
            bits = 3
        num_inputs = 2 * bits
        circuit = multiplierCircuit(bits)
    
    elif circuit_type == MULTIPLEXER:
        if bits == -1:
            bits = 3
        log_bits = bits
        num_inputs = 2**log_bits + log_bits
        circuit = multiplexerCircuit(bits)
    
    else:
        assert False
    
    bbs = {}
    additional_vars = []
    
    gates = circuit.getGates()
    gates = [g for g in gates if g.gate_type != Gate.NOT] # filter NOTs
    
    # randomly select blackboxes
    bbs_gates = random.sample(gates, num_bbs)
    
    # randomly select error gates
    error_gates = {}
    if error:
        gate = random.choice(gates)
        while gate in bbs_gates:
            gate = random.choice(gates)
        assert gate not in bbs_gates
        error_gate = random.choice(Gate.GATES)
        while error_gate == gate.gate_type:
            error_gate = random.choice(Gate.GATES)
        assert error_gate != gate.gate_type
        
        error_gates[gate] = error_gate
    
    # expand size of blackboxes if possible
    bbs_operands = {}
    for gate in bbs_gates:
        bbs_operands[gate] = copy.copy(gate.operands)
        size = 1
        while size < black_box_size:
            if not expandBlackBox(bbs_operands[gate], bbs_gates, error_gates):
                break
            size += 1
    
    encoded = []
    assumptions = []
    
    dot = set()
    for o in circuit.outputs:
        dot.update(o.toDot())
        o, assumptions_ = formulaFromCircuit(o, bbs_gates, bbs_operands, error_gates)
        encoded.append(o)
        assumptions += assumptions_
    
    temp = []
    for d in dot:
        temp.append(d)
    
    for blackbox in bbs_gates:
        temp.append('{}[shape=box,color=green][label="{}"];'.format(blackbox.__hash__(), blackbox.gate_type))
    for error_gate in error_gates:
        temp.append('{}[shape=box,color=red][label="{}/{}"];'.format(error_gate.__hash__(), error_gate.gate_type, error_gates[error_gate]))
    dot = 'digraph dependencies {{\n{}\n}}'.format('\n'.join(temp))
    
    return encoded, assumptions, dot

def expandBlackBox(operands, bbs_gates, error_gates):
    expansion_possible = False
    for i, operand in enumerate(operands):
        if not operand.isInput() and not operand in bbs_gates and not operand in error_gates:
            expansion_possible = True
            break
    
    if not expansion_possible:
        return False
    
    operand = operands[i]
    del operands[i]
    operands.extend(operand.operands)
    return True

def errorGateReachable(bbs_gates, error_gate):
    for blackbox in bbs_gates:
        if depthFirstSearch(blackbox, error_gate):
            return 1
        if depthFirstSearch(error_gate, blackbox):
            return 2
    return 0
    
def numErrorGateReachable(bbs_gates, error_gate):
    count = 0
    for blackbox in bbs_gates:
        if depthFirstSearch(blackbox, error_gate):
            count += 1
        elif depthFirstSearch(error_gate, blackbox):
            count += 1
    return count
    
def numInfluencedOutputs(outputs, error_gate):
    count = 0
    for output in outputs:
        if depthFirstSearch(output, error_gate):
            count += 1
    return count

def depthFirstSearch(source, goal):
    if source == goal:
        return True
    elif not source.isGate():
        return False
    else:
        for o in source.operands:
            if depthFirstSearch(o, goal):
                return True
        return False
    
def formulaFromCircuit(gate, bbs_gates, bbs_operands, error_gates):
    global num_y_vars, yvars, num_aux_vars, bbs_gates_processed, count_gates_processed
    
    if gate.isInput():
        o = gate.name
        if gate.name == '1': # TODO: Hack, there should be always at least one input
            o = '(x1 | !x1)'
        elif gate.name == '0':
            o = '(x1 & !x1)'
        return o, []
    
    elif gate in bbs_gates:
        assumptions = []
        
        if not gate in bbs_gates_processed:
            num_y_vars += 1
            yvar = 'y{}'.format(num_y_vars)
            bbs_gates_processed[gate] = yvar
            dependent = []
            for operand in bbs_operands[gate]:
                if operand.isInput():
                    if operand.name not in ('0','1'): # TODO: Hack, never add constant input to black box
                        dependent.append(operand.name)
                else:
                    num_aux_vars += 1
                    aux_var = 'z{}'.format(num_aux_vars)
                    dependent.append(aux_var)
                    formula, assumption = formulaFromCircuit(operand, bbs_gates, bbs_operands, error_gates)
                    assumptions += assumption
                    assumptions.append('({} <--> {})'.format(aux_var, formula))
            
            yvars[yvar] = dependent
            
        else:
            yvar = bbs_gates_processed[gate]
            
        return yvar, assumptions
        
    elif use_circuit_aux_vars and (gate.aux_var or gate in count_gates_processed):
        assumptions = []
        if not gate in count_gates_processed:
            num_aux_vars += 1
            aux_var = 'z{}'.format(num_aux_vars)
            gate.aux_var = False
            formula, assumption = formulaFromCircuit(gate, bbs_gates, bbs_operands, error_gates)
            assumptions += assumption
            assumptions.append('({} <--> {})'.format(aux_var, formula))
            count_gates_processed[gate] = aux_var
        else:
            aux_var = count_gates_processed[gate]
    
        return aux_var, assumptions
        
    elif gate.gate_type == Gate.NOT:
        formula, assumption = formulaFromCircuit(gate.operands[0], bbs_gates, bbs_operands, error_gates)
        return '!{}'.format(formula), assumption
    else:
        lhs, assumption1 = formulaFromCircuit(gate.operands[0], bbs_gates, bbs_operands, error_gates)
        rhs, assumption2 = formulaFromCircuit(gate.operands[1], bbs_gates, bbs_operands, error_gates)
        
        gate_function = gate.gate_type
        if gate in error_gates:
            gate_function = error_gates[gate]
        
        return gate_translation[gate_function].format(lhs, rhs), assumption1 + assumption2

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('circuit', metavar='circuit', choices=circuits,
                        help='{{{}}}'.format(','.join(circuits)))
    parser.add_argument('blackboxes', type=int,
                        help='Number of black boxes')
    parser.add_argument('instances', type=int,
                        help='Number of instances to create')
    parser.add_argument('directory',
                        help='Directory to write the instances')
    parser.add_argument('bits', type=int, nargs='?',
                        help='Number of input bits for the circuit')
    parser.add_argument('blackbox_size', type=int, nargs='?',
                        help='Size of the black boxes (default 1 gate)')
    
    args = parser.parse_args()
    
    circuit = args.circuit
    bbs = args.blackboxes
    instances = args.instances
    directory = args.directory
    if args.bits:
        bits = args.bits
    if args.blackbox_size:
        black_box_size = args.blackbox_size
    
    random.seed(0) # make deterministic
    
    if not os.path.exists(directory):
        os.mkdir(directory)
    
    outputs_spec, assumptions, _ = buildCircuit(circuit, 0)
    #assert not assumptions
    orig_num_aux_vars = num_aux_vars
    orig_assumptions = assumptions
    
    inputs = ['x{}'.format(i+1) for i in xrange(num_inputs)]
    
    print("Create {} instances with {} black boxes".format(instances, bbs))
    for i in xrange(instances):
        
        # reset global vars
        num_y_vars = 0
        yvars = {}
        num_aux_vars = orig_num_aux_vars
        bbs_gates_processed = {}
        count_gates_processed = {}
        Gate.cache = {}
        
        if i == 0:
            outputs, assumptions, dot = buildCircuit(circuit, bbs)
        else:
            outputs, assumptions, dot = buildCircuit(circuit, bbs, error=True)
        
        dot_file_name = 'pec_{}_{}_{}.dot'.format(circuit, bbs, i)
        open(os.path.join(directory, dot_file_name), 'w').write(dot)
        
        assumptions = orig_assumptions + assumptions
        
        assert num_y_vars == bbs
        
        file_name = 'pec_{}_{}_{}.dqbf'.format(circuit, bbs, i)
        
        file_handle = open(os.path.join(directory, file_name), 'w')
        
        file_handle.write('A {} : '.format(' '.join(inputs + ['z{}'.format(z) for z in xrange(1, num_aux_vars+1)])))
        for y in yvars:
            file_handle.write('E{{{}}} {} : '.format(' '.join(yvars[y]), y))
        
        specification = '({})'.format(' & '.join(['({} <--> {})'.format(a,b) for a,b in zip(outputs, outputs_spec)]))
        
        if assumptions:
            assumptions = '({})'.format(' | '.join(['!{}'.format(a) for a in assumptions]))    
            file_handle.write('({} | {})'.format(assumptions, specification))
        else: 
            file_handle.write(specification)
        
        file_handle.close()
        
    