# coding=utf-8

"""Library of syntax tree modifications."""

__author__ = "Leander Tentrup <tentrup@cs.uni-saarland.de>"
__copyright__ = "Copyright 2014, Saarland University"

from parse import *
from lex import *

'''
  Negates a formula
'''
class Negator(object):
    
    def __init__(self):
        self.formula = None
    
    def visitQuantifier(self, q):
        
        mapping = {TokenType.E : TokenType.A, TokenType.A : TokenType.E}
        q.qtype = mapping[q.qtype]
            
        q.operands[0].accept(self) # result in self.formula
        
        self.formula = q
        
    def visitIdent(self, ident):
        assert False
        
    def visitProposition(self, prop):
        self.formula = UnaryOp(TokenType.NOT, prop)
    
    def visitUnaryOp(self, unop):
        mapping = {
            TokenType.GLOBALLY : TokenType.FINALLY,
            TokenType.FINALLY  : TokenType.GLOBALLY
        }
        
        if unop.optype == TokenType.NOT:
            self.formula = unop.operands[0]
            
        elif unop.optype == TokenType.NEXT:
            unop.operands[0].accept(self)
            unop.operands[0] = self.formula
            self.formula = unop
            
        elif unop.optype in mapping:
            unop.optype = mapping[unop.optype]
            unop.operands[0].accept(self)
            unop.operands[0] = self.formula
            self.formula = unop
        else:
            assert False


    def visitBinaryOp(self, binop):
        mapping = {
            TokenType.AND     : TokenType.OR,
            TokenType.OR      : TokenType.AND,
            TokenType.UNTIL   : TokenType.RELEASE,
            TokenType.RELEASE : TokenType.UNTIL
        }
        
        if binop.optype in mapping:
            
            assert binop.optype not in [TokenType.UNTIL,TokenType.RELEASE] or len(binop.operands) == 2
            
            binop.optype = mapping[binop.optype]
            
            for i in xrange(len(binop.operands)):            
                binop.operands[i].accept(self)
                binop.operands[i] = self.formula
            
            self.formula = binop
                
        elif binop.optype == TokenType.IMPL:
            
            binop.optype = TokenType.AND
            binop.operands[1].accept(self)
            binop.operands[1] = self.formula
            
            self.formula = binop
            
        elif binop.optype == TokenType.EQUIV:
            self.formula = UnaryOp(TokenType.NOT, binop)
        elif binop.optype == TokenType.XOR:
            self.formula = UnaryOp(TokenType.NOT, binop)

        else:
            assert False
    
    def __str__(self):
        return str(self.formula)
        
'''
  Flattens the tree structure, combines consecutive ANDs and ORs
'''
class Flattening(object):
    def __init__(self):
        self.formula = None
         
    def visitQuantifier(self, q):
        q.operands[0].accept(self) # result in self.formula
        q.operands[0] = self.formula
        self.formula = q
         
    def visitIdent(self, ident):
        assert False
         
    def visitProposition(self, prop):
        self.formula = prop
     
    def visitUnaryOp(self, unop):
        unop.operands[0].accept(self)
        unop.operands[0] = self.formula
        self.formula = unop
     
    def visitBinaryOp(self, binop):
        if binop.optype in [TokenType.AND, TokenType.OR]:
            i = 0
            while i < len(binop.operands):
                try:
                    if binop.operands[i].optype == binop.optype:
                        binop.operands = binop.operands[:i] + binop.operands[i].operands + binop.operands[i+1:]
                        i -= 1
                except AttributeError:
                    pass
                i += 1
             
            j = 0
            while j < len(binop.operands):
                binop.operands[j].accept(self)
                binop.operands[j] = self.formula
                j += 1
             
            self.formula = binop
        
        elif binop.optype in TokenType.IMPL:
            binop.operands[0].accept(self)
            binop.operands[0] = self.formula
            binop.operands[1].accept(self)
            binop.operands[1] = self.formula
            
            self.formula = binop    
        
        else:
            self.formula = binop
 
    def __str__(self):
        return str(self.formula)

'''
  Removes syntactical aliase from the syntax tree
'''
class Normalizer(object):
    
    def visitQuantifier(self, q):
        q.operands[0].accept(self)
         
    def visitIdent(self, ident):
        assert False
         
    def visitProposition(self, prop):
        pass
     
    def visitUnaryOp(self, unop):
        unop.operands[0].accept(self)
        if unop.optype in TokenType.ALIAS:
            unop.optype = TokenType.ALIAS[unop.optype]
     
    def visitBinaryOp(self, binop):
        for o in binop.operands:
            o.accept(self)
        
        if binop.optype in TokenType.ALIAS:
            binop.optype = TokenType.ALIAS[binop.optype]

'''
  Removes identifier from quantifier
'''
class RemoveIdentifier(object):
    
    def visitQuantifier(self, q):
        q.operands[0].accept(self)
        
        q.variables = [v.name for v in q.variables]
         
    def visitIdent(self, ident):
        assert False
         
    def visitProposition(self, prop):
        pass
     
    def visitUnaryOp(self, unop):
        pass
     
    def visitBinaryOp(self, binop):
        pass

'''
  Copies the syntax tree
'''
class Copier(object):
    
    def __init__(self, name_mapping):
        self.formula = None
        self.name_mapping = name_mapping
    
    def visitQuantifier(self, q):
        assert False
         
    def visitIdent(self, ident):
        assert False
         
    def visitProposition(self, prop):
        self.formula = Proposition(self.name_mapping(prop.name))
     
    def visitUnaryOp(self, unop):
        unop.operands[0].accept(self)
        self.formula = UnaryOp(unop.optype, self.formula)
     
    def visitBinaryOp(self, binop):
        operands = []
        for o in binop.operands:
            o.accept(self)
            operands.append(self.formula)
        
        self.formula = BinOp(binop.optype, operands=operands)